@@ -1,21 +1,16 @@
 import asyncio
 import logging
 import os
-import random
-import string
 import subprocess
 import sys
 from collections import defaultdict
 from datetime import datetime
 
-import zmq.asyncio
+import aiohttp
+import dyne.org.funtoo.powerbus as powerbus
 from dbus_next import Message, InterfaceNotFoundError, BusType
-
-import dyne.org.funtoo.powerbus.zmq_bus as zmq_bus
 from dbus_next.aio import MessageBus
 
-from subpop.hub import DeferredHubProperty
-
 
 async def get_bus_names():
 	"""
@@ -147,40 +142,52 @@ async def prepare_for_sleep():
 
 
 class FuntooIdleClient(object):
-	def __init__(self, endpoint=zmq_bus.BUS_PATH, identity=None):
-		self.endpoint = endpoint
-		self.ctx = zmq.asyncio.Context()
-		self.client: zmq.asyncio.Socket = self.ctx.socket(zmq.DEALER)
+	def __init__(self):
+		self.last_msg = datetime.utcnow()
+		self.ws = None
+		self.running = True
 
-		# https://stackoverflow.com/questions/60324740/0mq-can-you-drop-a-message-after-a-timeout-in-a-req-rep-pattern
-		# These lines, combined with a msg.send(flags=zmq.NOWAIT), will result in our client not queueing up messages
-		# to the daemon if the daemon is not available. Messages will just be discarded. This prevents a flood of stale
-		# messages being sent to the daemon when it starts up.
+	async def connect(self):
+		while True:
+			try:
+				logging.warning("Connecting to remote.")
+				conn = aiohttp.UnixConnector(path=powerbus.socket.POWERBUS_PATH)
+				async with aiohttp.ClientSession(connector=conn) as session:
+					async with session.ws_connect("http://sock/ws") as self.ws:
 
-		self.client.setsockopt(zmq.LINGER, 0)
-		self.client.setsockopt(zmq.IMMEDIATE, 1)
-		self.client.setsockopt(zmq.CONFLATE, 1)
-		self.client.setsockopt(zmq.RCVTIMEO, 1000)
-		if identity is None:
-			self.client.setsockopt(
-				zmq.IDENTITY, ("".join(random.choice(string.ascii_uppercase + string.digits) for _ in range(16)).encode("ascii"))
-			)
-		else:
-			self.client.setsockopt(zmq.IDENTITY, identity.encode("utf-8"))
-		logging.info("Connecting to " + self.endpoint)
-		self.client.connect(self.endpoint)
-		self.ping_event = asyncio.Event()
+						logging.warning("Got websocket.")
+						while True:
+							await self.sendinfo()
+							try:
+								msg = await asyncio.wait_for(self.ws.receive(), timeout=2)
+							except asyncio.futures.TimeoutError:
+								logging.warning("Receive timeout.")
+								continue
+							self.last_msg = datetime.utcnow()
+							if msg.type == aiohttp.WSMsgType.TEXT:
+								msg_obj = await self.on_recv(msg)
+								if msg_obj is not None:
+									await msg_obj.send(self.ws)
+							elif msg.type == aiohttp.WSMsgType.ERROR:
+								raise ConnectionResetError()
+							elif msg.type == aiohttp.WSMsgType.PING:
+								sys.stdout.write(".")
+								sys.stdout.flush()
+			except (aiohttp.ClientConnectionError, ConnectionResetError) as e:
+				print(e)
+				print("reconnecting in 2 seconds...")
+				await asyncio.sleep(2)
 
-	async def on_recv(self, msg):
-		msg_obj = zmq_bus.PowerBusMessage.from_msg(msg)
-		print("Got", msg_obj)
+	async def on_recv(self, msg) -> powerbus.socket.PowerBusMessage:
+		msg_obj = powerbus.socket.PowerBusMessage.from_msg(msg)
+		self.last_msg = datetime.utcnow()
 		action_method_name = "action_" + msg_obj.action.replace("-", "_")
 		action_method = getattr(self, action_method_name, None)
 		if action_method:
 			resp_obj = await action_method(msg_obj)
 			if resp_obj:
 				# Note that this is not really a response. This is a new message "info" or "req" going out...
-				await resp_obj.send(self.client)
+				await resp_obj.send(self.ws)
 			else:
 				logging.error(f"I really did not expect this: {msg}")
 		else:
@@ -189,28 +196,16 @@ class FuntooIdleClient(object):
 	async def action_prepare_for_sleep(self, msg_obj):
 		logging.warning("Preparing for sleep...")
 		await prepare_for_sleep()
-		return zmq_bus.PowerBusMessage(action=msg_obj.action, status="success")
+		self.running = False
+		return powerbus.socket.PowerBusMessage(action=msg_obj.action, status="success")
 
 	async def start(self):
-		await hub.add_property("SESSION_BUS", DeferredHubProperty(lambda: MessageBus(bus_type=BusType.SESSION).connect()))
-		asyncio.create_task(self.ping_task())
-		# await on this -- it will never end -- so this method will never end -- which is what we want:
-		await self.recv_task()
+		hub.SESSION_BUS = await MessageBus(bus_type=BusType.SESSION).connect()
+		await self.connect()
 
-	async def recv_task(self):
-		while True:
-			try:
-				msg = await self.client.recv_multipart()
-				await self.on_recv(msg)
-			except zmq.ZMQError:
-				# Likely no message.
-				pass
-
-	async def ping(self):
-		sys.stdout.write(".")
-		sys.stdout.flush()
+	async def sendinfo(self):
 		out = subprocess.getoutput("/usr/bin/funtoo-xidle")
-		out_msg = zmq_bus.PowerBusMessage(
+		out_msg = powerbus.socket.PowerBusMessage(
 			action="ping",
 			idle=int(out),
 			time=datetime.utcnow(),
@@ -219,9 +214,4 @@ class FuntooIdleClient(object):
 			inhibitors=await get_session_inhibitors(),
 		)
 		print("idle =", out_msg.idle)
-		await out_msg.send(self.client)
-
-	async def ping_task(self):
-		while True:
-			await self.ping()
-			await asyncio.sleep(1)
+		await out_msg.send(self.ws)
