@@ -6,14 +6,14 @@ import os
 import subprocess
 import sys
 from datetime import datetime, timezone, timedelta
+from enum import Enum
 
+import aiohttp
 import colorama
-import zmq.asyncio
+from aiohttp import web
 from dbus_next import Message, BusType
 from dbus_next.aio import MessageBus
 
-from subpop.hub import DeferredHubProperty
-
 LID_CLOSED_EVENT = asyncio.Event()
 LID_LAST_CLOSED = None
 LID_LAST_OPENED = None
@@ -145,6 +145,12 @@ async def sleep_inhibited():
 	return False
 
 
+class SleepState(Enum):
+	ACTIVE = 1
+	SLEEPING = 2
+	ACCEL_SLEEP = 3
+
+
 async def setup_lid_callback():
 	"""
 	This method will set up a realtime callback that gets called when the system's 'lid' is opened or closed, so
@@ -166,11 +172,11 @@ async def setup_lid_callback():
 			if my_dict["LidIsClosed"].value:
 				print(str(my_dict["LidIsClosed"]))
 				logging.warning("Lid is closed!")
-				LID_LAST_CLOSED = datetime.utcnow()
+				LID_LAST_CLOSED = get_utc_time()
 				LID_CLOSED_EVENT.set()
 				LID_CLOSED_EVENT.clear()
 			else:
-				LID_LAST_OPENED = datetime.utcnow()
+				LID_LAST_OPENED = get_utc_time()
 
 	bus_name = "org.freedesktop.UPower"
 	path = "/" + bus_name.replace(".", "/")
@@ -183,36 +189,6 @@ async def setup_lid_callback():
 class PowerBusDaemon(object):
 	def __init__(self):
 
-		if not self.is_root:
-			raise PermissionError()
-
-		if not os.path.isdir("/run"):
-			raise FileNotFoundError("/run not found and required by funtoo-powerbus.")
-		os.umask(0o022)
-		if not os.path.exists("/run/funtoo"):
-			os.mkdir("/run/funtoo", 0o700)
-
-		# Give us exclusive access to this directory while we set it up:
-		os.chown("/run/funtoo", 0, 0)
-		os.chmod("/run/funtoo", 0o700)
-		try:
-			put_file("/run/funtoo/powerbus.pid", str(os.getpid()))
-		except PermissionError:
-			raise PermissionError("Unable to write PID file at /runt/funtoo/powerbus.pid.")
-
-		os.umask(0o000)
-		# ZMQ Setup:
-		self.bind_addr = powerbus.zmq_bus.BUS_PATH
-		self.ctx = zmq.asyncio.Context()
-		self.identities = {}
-		self.server: zmq.asyncio.Socket = self.ctx.socket(zmq.ROUTER)
-		self.server.bind(self.bind_addr)
-
-		# Now that all files (including powerbus-socket) are in place -- relax permissions.
-		os.chmod("/run/funtoo", 0o755)
-		os.umask(0o022)
-		logging.debug(f"Listening for new client connections at {self.bind_addr}")
-
 		# PowerBus Internals Setup:
 		self.suspend_index = 0
 		self.idle_delay_sec = 90
@@ -227,31 +203,101 @@ class PowerBusDaemon(object):
 		self.clients = {}
 		self.epoch = datetime.utcfromtimestamp(0)
 		self.stale_msg_interval = timedelta(milliseconds=500)
+		self.last_zero = get_utc_time()
+		self.state = SleepState.ACTIVE
+		self.app = None
+		self.server = None
+		self.runner = None
+		if not self.is_root:
+			raise PermissionError()
+
+	async def stop_server(self):
+		await self.runner.cleanup()
+
+	async def start_server(self):
+		"""
+		This method starts the server as a background Task, and returns.
+		"""
+
+		async def websocket_handler(request):
+			logging.warning("Got connection!")
+			ws = web.WebSocketResponse()
+			await ws.prepare(request)
+
+			async for msg in ws:
+				logging.warning(f"Got message! {msg.data}")
+				if msg.type == aiohttp.WSMsgType.TEXT:
+					msg_obj = await self.on_recv(ws, msg)
+					if msg_obj is not None:
+						await msg_obj.send(ws)
+				elif msg.type == aiohttp.WSMsgType.ERROR:
+					logging.warning(f"ws connection closed with exception {ws.exception()}")
+					self.remove_client(ws)
+					break
+
+			self.remove_client(ws)
+			# connection closed
+			return ws
+
+		if not self.is_root:
+			raise PermissionError()
+
+		if not os.path.isdir("/run"):
+			raise FileNotFoundError("/run not found and required by funtoo-powerbus.")
+		os.umask(0o022)
+		if not os.path.exists("/run/funtoo"):
+			os.mkdir("/run/funtoo", 0o755)
+		os.chmod("/run/funtoo", 0o755)
+		os.chown("/run/funtoo", 0, 0)
+		try:
+			put_file("/run/funtoo/powerbus.pid", str(os.getpid()))
+		except PermissionError:
+			raise PermissionError("Unable to write PID file at /run/funtoo/powerbus.pid.")
+
+		self.app = web.Application()
+		self.app.add_routes([web.get("/ws", websocket_handler)])
+		self.runner = web.AppRunner(self.app)
+		await self.runner.setup()
+		os.umask(0o000)
+		site = web.UnixSite(self.runner, path=powerbus.socket.POWERBUS_PATH)
+		await site.start()
+		os.umask(0o022)
+		logging.debug(f"Listening for new client connections at {powerbus.socket.POWERBUS_PATH}")
 
 	@property
 	def is_root(self):
 		return os.geteuid() == 0
 
 	def go_to_sleep(self):
+		self.last_zero = None
+		self.state = SleepState.SLEEPING
 		subprocess.getoutput("echo -n mem > /sys/power/state")
 
-	async def send_message_to_all_active_clients(self, msg_obj: powerbus.zmq_bus.PowerBusMessage):
-		print("Going to possibly send messages...")
+	async def send_ping_to_all_active_clients(self):
 		cur_time = get_utc_time()
-		for client_key, orig_msg_obj in self.clients.items():
+		for client_key, datums in self.clients.items():
+			ws, orig_msg_obj = datums
+			if cur_time - orig_msg_obj.time <= self.client_stale_timeout:
+				await ws.ping()
+
+	async def send_message_to_all_active_clients(self, msg_obj: powerbus.socket.PowerBusMessage):
+		cur_time = get_utc_time()
+		for client_key, datums in self.clients.items():
+			ws, orig_msg_obj = datums
 			if cur_time - orig_msg_obj.time <= self.client_stale_timeout:
 				print("Sending message", msg_obj)
-				await msg_obj.send(self.server, identity=orig_msg_obj.identity)
+				await msg_obj.send(ws)
 
 	async def tell_clients_to_prepare_for_sleep(self):
-		msg_obj = powerbus.zmq_bus.PowerBusMessage(action="prepare-for-sleep")
+		msg_obj = powerbus.socket.PowerBusMessage(action="prepare-for-sleep")
 		await self.send_message_to_all_active_clients(msg_obj)
 
 	@property
 	def num_clients(self):
 		num_fresh_clients = 0
 		cur_time = get_utc_time()
-		for client_key, msg_obj in self.clients.items():
+		for client_key, datums in self.clients.items():
+			ws, msg_obj = datums
 			if cur_time - msg_obj.time <= self.client_stale_timeout:
 				num_fresh_clients += 1
 		return num_fresh_clients
@@ -260,9 +306,8 @@ class PowerBusDaemon(object):
 
 		"""
 		This method will look at all non-stale clients connected, and look at their idle times. If there is at least
-		one actively communicating client, and all clients' idle times are greater than ``self.idle_sec``, then we
-		have passed our first test in determining if we should sleep, and we will return True. Otherwise, this method
-		will return False.
+		one actively communicating client, and all clients' idle times are sufficient, then we will return True.
+		Otherwise False.
 
 		This doesn't mean we will definitely go to sleep, because sleep could be inhibited via logind. But we can now
 		proceed to see if it's bed time.
@@ -270,39 +315,79 @@ class PowerBusDaemon(object):
 		:return: Whether or not the system is sufficiently idle to go to sleep.
 		:rtype: bool
 		"""
+
+		# This means we are recovering from sleep:
+
+		if self.last_zero is None or self.state == SleepState.SLEEPING:
+			return False
+
 		num_fresh_clients = 0
 		num_idle_clients = 0
 		cur_time = get_utc_time()
-		for client_key, msg_obj in self.clients.items():
-			if cur_time - msg_obj.time <= self.client_stale_timeout:
-				num_fresh_clients += 1
-				if msg_obj.idle > timeout_ms:
-					num_idle_clients += 1
 
-		if num_fresh_clients > 0 and num_idle_clients == num_fresh_clients:
+		timeout_delta = timedelta(milliseconds=timeout_ms)
+
+		# Our first test -- we must be at least timeout_ms away from our self.last_zero point. This timer gets reset
+		# when certain special events happen (daemon start, resume from sleep...)
+
+		if cur_time - self.last_zero < timeout_delta:
+			# Not enough time has passed.
+			return False
+
+		# Now let's look at each client reporting idle data back to us.
+
+		for client_key, datums in self.clients.items():
+			ws, msg_obj = datums
+			# We expect the client's last idle message to be at least somewhat fresh. Otherwise we'll skip it.
+
+			time_offset = cur_time - msg_obj.time
+
+			if time_offset > self.client_stale_timeout:
+				continue
+
+			num_fresh_clients += 1
+
+			# Now, let's see if this client is sufficiently idle, according to our standards. msg_obj.idle stores the idle time
+			# of the user, in milliseconds.
+
+			idle = timedelta(milliseconds=msg_obj.idle)
+
+			# Correct for the time passed since we received this message:
+
+			idle += time_offset
+
+			if idle > timeout_delta:
+				num_idle_clients += 1
+
+		if 0 < num_fresh_clients == num_idle_clients:
 			return True
 		else:
 			return False
 
 	def session_sleep_inhibited(self):
 		user_inhibited = {}
-		for client_key, msg_obj in self.clients.items():
+		for client_key, datums in self.clients.items():
+			ws, msg_obj = datums
 			if len(msg_obj.inhibitors):
 				user_inhibited[client_key] = msg_obj.inhibitors
 		return len(user_inhibited) > 0, user_inhibited
 
 	async def start(self):
-		await hub.add_property("SYSTEM_BUS", DeferredHubProperty(lambda: MessageBus(bus_type=BusType.SYSTEM).connect()))
+		hub.SYSTEM_BUS = await MessageBus(bus_type=BusType.SYSTEM).connect()
 		self.has_lid = await get_upower_property("lid_is_present")
 		if self.has_lid:
 			await setup_lid_callback()
+		asyncio.create_task(self.state_manager())
 		asyncio.create_task(self.sleep_handler())
 		asyncio.create_task(self.lid_event_monitor_task())
 		asyncio.create_task(self.status_display())
 		asyncio.create_task(self.stale_cleanup_task())
+		await self.start_server()
 		while True:
-			msg = await self.server.recv_multipart()
-			self.on_recv(msg)
+			# ASCII heartbeat
+			sys.stdout.write(".")
+			sys.stdout.flush()
+			await asyncio.sleep(1)
 
 	def get_epoch_seconds(self, dt):
 		return (dt - self.epoch).total_seconds()
@@ -363,10 +448,11 @@ class PowerBusDaemon(object):
 			await asyncio.sleep(5)
 			cur_time = get_utc_time()
 			new_clients = {}
-			for client_key, msg_obj in self.clients.items():
+			for client_key, datums in self.clients.items():
+				ws, msg_obj = datums
 				if cur_time - msg_obj.time > self.client_stale_timeout:
 					continue
-				new_clients[client_key] = msg_obj
+				new_clients[client_key] = datums
 			self.clients = new_clients
 
 	async def lid_event_monitor_task(self):
@@ -378,9 +464,30 @@ class PowerBusDaemon(object):
 		while True:
 			await LID_CLOSED_EVENT.wait()
 			if not on_ac_power():
-				self.last_maybe_accelerated_sleep = datetime.utcnow()
+				self.last_maybe_accelerated_sleep = get_utc_time()
 				self.lid_closed_accelerated_sleep = True
 
+	async def state_manager(self):
+		"""
+		This state manager is designed to pull the system out of a SLEEPING state on resume.
+
+		When the system goes to sleep, self.state == SleepState.SLEEPING and self.last_zero is None.
+		After one second, in this state, we will set self.state to SleepState.ACTIVE. Then after another
+		second, self.last_zero will be set to the current time.
+
+		Since we set the SLEEPING state and the last zero of None right before sleep, and the system
+		will sleep after milliseconds, and this will take 2 seconds to go back to fully active mode,
+		we will become fully active only AFTER we have resumed. And if for some reason going to sleep
+		fails, this pulls us back into an active state in 2 seconds.
+		"""
+		while True:
+			await asyncio.sleep(1)
+			if self.state == SleepState.SLEEPING:
+				self.state = SleepState.ACTIVE
+			elif self.state == SleepState.ACTIVE:
+				if self.last_zero is None:
+					self.last_zero = get_utc_time()
+
 	async def sleep_handler(self):
 		"""
 		This method actually sees if we should go to sleep -- now -- and if so, will put the system to sleep.
@@ -389,7 +496,12 @@ class PowerBusDaemon(object):
 		and modified or enhanced if desired.
 		"""
 		while True:
-			await self.send_message_to_all_active_clients(powerbus.zmq_bus.PowerBusMessage({"action": "ping"}))
+			# We are not fully awake yet -- so don't consider sleeping.
+			if self.state == SleepState.SLEEPING or self.last_zero is None:
+				await asyncio.sleep(2)
+				continue
+
+			await self.send_ping_to_all_active_clients()
 			await asyncio.sleep(2)
 
 			# We use timeout_ms set to a value to indicate that we have determined that we should consider sleep
@@ -441,10 +553,10 @@ class PowerBusDaemon(object):
 				# Therefore, update the timestamp for the last accelerated sleep, and also turn off accel.
 				# sleep mode -- it can only be active for one shot, then it gets disabled until the lid is
 				# closed again.
-				self.last_accelerated_sleep = datetime.utcnow()
+				self.last_accelerated_sleep = get_utc_time()
 				self.lid_closed_accelerated_sleep = False
 			else:
-				self.last_sleep_attempt = datetime.utcnow()
+				self.last_sleep_attempt = get_utc_time()
 
 			await self.tell_clients_to_prepare_for_sleep()
 			await asyncio.sleep(0.25)
@@ -468,6 +580,8 @@ class PowerBusDaemon(object):
 					f" * {colorama.Fore.CYAN}Status:{colorama.Style.RESET_ALL} Will not naturally sleep. Reason: {self.color(reason)}"
 				)
 			print(f" * Daemon start time:  {self.color(self.utc_to_local(self.daemon_start))}")
+			print(f" * Last zero:          {self.color(self.utc_to_local(self.last_zero))}")
+			print(f" * State:              {self.color(self.state)}")
 			if self.has_lid:
 				print(f" * Lid Last Closed:    {self.color(self.utc_to_local(LID_LAST_CLOSED))}")
 				print(f" * Lid Last Opened:    {self.color(self.utc_to_local(LID_LAST_OPENED))}")
@@ -503,13 +617,15 @@ class PowerBusDaemon(object):
 					for app_name, extras in inhibitor_dict.items():
 						print(f"     \\ {app_name} : {extras}")
 			print(colorama.Style.RESET_ALL)
-			await asyncio.sleep(2)
+			await asyncio.sleep(5)
+
+	async def on_recv(self, ws, msg_str):
+		msg_obj = powerbus.socket.PowerBusMessage.from_msg(msg_str)
+		for req_key in ["USER", "DISPLAY", "time", "idle", "inhibitors"]:
+			if msg_obj.get(req_key, None) is None:
+				logging.warning(f"Received invalid message! Missing: {req_key}")
+				return
 
-	def on_recv(self, msg):
-		# First, receive and inspect message:
-		zmq_identity = msg[0]
-		logging.info("Received: %s" % msg)
-		msg_obj = powerbus.zmq_bus.PowerBusMessage.from_msg(msg[1])
 		msg_time = msg_obj.time
 		cur_time = get_utc_time()
 
@@ -523,15 +639,30 @@ class PowerBusDaemon(object):
 			print("OLD message!")
 			return
 
-		existing_keys = ["USER", "DISPLAY"]
 		client_key = set()
-		for key in existing_keys:
-			if key in msg_obj:
-				client_key.add((key, msg_obj[key]))
-		# Record identity for sending messages back:
-		msg_obj.identity = zmq_identity
+		for key in ["USER", "DISPLAY"]:
+			client_key.add((key, msg_obj[key]))
 
-		self.clients[frozenset(client_key)] = msg_obj
+		self.clients[client_key_from_msg(msg_obj)] = (ws, msg_obj)
+
+	def remove_client(self, ws):
+		"""
+		Remove active client connection identified by live WebSocket object.
+		"""
+		client_key = None
+		for cur_client, datums in self.clients.items():
+			if id(datums[0]) == id(ws):
+				client_key = cur_client
+				break
+		if client_key:
+			del self.clients[client_key]
+
+
+def client_key_from_msg(msg_obj):
+	client_key = set()
+	for key in ["USER", "DISPLAY"]:
+		client_key.add((key, msg_obj[key]))
+	return frozenset(client_key)
 
 
 def client_key_to_str(client_key):
